--- ethereal_orig/packet-tcp.c	2004-05-15 02:41:26.000000000 +0200
+++ ethereal_new/packet-tcp.c	2004-05-16 10:27:33.843750000 +0200
@@ -248,6 +248,7 @@
 	guint32 base_seq2;
 	gint16 win_scale1;
 	gint16 win_scale2;
+	guint32 server_port;
 	guint32 ack1, ack2;
 	guint32 ack1_frame, ack2_frame;
 	nstime_t ack1_time, ack2_time;
@@ -259,7 +260,7 @@
 	struct tcp_next_pdu *pdu_seq1;
 	struct tcp_next_pdu *pdu_seq2;
 };
-
+static gboolean tcp_mark_server_port = TRUE;
 
 static GMemChunk *tcp_next_pdu_chunk = NULL;
 static int tcp_next_pdu_count = 20;
@@ -290,6 +291,7 @@
 	if(!tcpd){
 		/* No no such data yet. Allocate and init it */
 		tcpd=g_mem_chunk_alloc(tcp_analysis_chunk);
+		tcpd->server_port = 0;
 		tcpd->ual1=NULL;
 		tcpd->base_seq1=0;
 		tcpd->win_scale1=-1;
@@ -316,6 +318,27 @@
 	return tcpd;
 }
 
+/* Mark server port for more reliable subdissector search based on TCP port. 
+   Since this feature implies to look for the conversation state for each packet,
+   it can be disabled with a preference */
+static void
+mark_server_port(packet_info *pinfo, guint8 th_flags)
+{
+    struct tcp_analysis *tcpd=NULL;
+    tcpd=get_tcp_conversation_data(pinfo);
+    if ((th_flags & TH_SYN) && !(th_flags & TH_ACK))
+    {
+      /* If this packet is a TCP SYN, store the destination port as server port */
+      tcpd->server_port  = pinfo->destport;
+      pinfo->server_port = pinfo->destport;
+    }
+    else
+    {
+      /* Else restore the server port as stored in the conversation from the TCP SYN */
+      pinfo->server_port = tcpd->server_port;  	  	
+    }
+}
+
 /* This function is called from the tcp analysis code to provide
    clues on how the seq and ack numbers are changed.
    To prevent the next_pdu lists from growing uncontrollable in size we
@@ -2356,39 +2379,48 @@
     }
   }
 
-  /* Do lookups with the subdissector table.
-     We try the port number with the lower value first, followed by the
-     port number with the higher value.  This means that, for packets
-     where a dissector is registered for *both* port numbers:
-
-	1) we pick the same dissector for traffic going in both directions;
-
-	2) we prefer the port number that's more likely to be the right
-	   one (as that prefers well-known ports to reserved ports);
-
-     although there is, of course, no guarantee that any such strategy
-     will always pick the right port number.
-
-     XXX - we ignore port numbers of 0, as some dissectors use a port
-     number of 0 to disable the port. */
-  if (src_port > dst_port) {
-    low_port = dst_port;
-    high_port = src_port;
-  } else {
-    low_port = src_port;
-    high_port = dst_port;
-  }
-  if (low_port != 0 &&
-      dissector_try_port(subdissector_table, low_port, next_tvb, pinfo, tree)){
-    pinfo->want_pdu_tracking -= !!(pinfo->want_pdu_tracking);
-    return TRUE;
+  /* Do lookups with the subdissector table. 
+     Check if the server port is known from a previously seen TCP SYN */
+  if (pinfo->server_port != 0)
+  {
+    if (dissector_try_port(subdissector_table, pinfo->server_port, next_tvb, pinfo, tree))
+	    return TRUE;
   }
-  if (high_port != 0 &&
-      dissector_try_port(subdissector_table, high_port, next_tvb, pinfo, tree)){
-    pinfo->want_pdu_tracking -= !!(pinfo->want_pdu_tracking);
-    return TRUE;
+  else
+  {     
+    /* We try the port number with the lower value first, followed by the
+       port number with the higher value.  This means that, for packets
+       where a dissector is registered for *both* port numbers:
+
+  	   1) we pick the same dissector for traffic going in both directions;
+
+	   2) we prefer the port number that's more likely to be the right
+	      one (as that prefers well-known ports to reserved ports);
+
+       although there is, of course, no guarantee that any such strategy
+       will always pick the right port number.
+
+       XXX - we ignore port numbers of 0, as some dissectors use a port
+       number of 0 to disable the port. */
+    if (src_port > dst_port) {
+      low_port = dst_port;
+      high_port = src_port;
+    } else {
+      low_port = src_port;
+      high_port = dst_port;
+    }
+    if (low_port != 0 &&
+        dissector_try_port(subdissector_table, low_port, next_tvb, pinfo, tree)){
+      pinfo->want_pdu_tracking -= !!(pinfo->want_pdu_tracking);
+      return TRUE;
+    }
+    if (high_port != 0 &&
+        dissector_try_port(subdissector_table, high_port, next_tvb, pinfo, tree)){
+      pinfo->want_pdu_tracking -= !!(pinfo->want_pdu_tracking);
+      return TRUE;
+    }
   }
-
+  
   if (!try_heuristic_first) {
     /* do lookup with the heuristic subdissector table */
     if (dissector_try_heuristic(heur_subdissector_list, next_tvb, pinfo, tree)){
@@ -2590,6 +2622,11 @@
   tcph->th_win = tvb_get_ntohs(tvb, offset + 14);
   tcph->th_hlen = hi_nibble(th_off_x2) * 4;  /* TCP header length, in bytes */
 
+  if (tcp_mark_server_port)
+  {
+  	mark_server_port(pinfo, tcph->th_flags);
+  }
+
   /*
    * If we've been handed an IP fragment, we don't know how big the TCP
    * segment is, so don't do anything that requires that we know that.
@@ -3229,6 +3266,11 @@
 	    "Try heuristic sub-dissectors first",
 	    "Try to decode a packet using an heuristic sub-dissector before using a sub-dissector registered to a specific port",
 	    &try_heuristic_first);
+	prefs_register_bool_preference(tcp_module, "server_port",
+	    "Mark server port",
+	    "Mark server port for each packet if the TCP SYN is part of the conversation. "
+	    "This option makes the sub-dissector search based on the TCP port more reliable.",
+	    &tcp_mark_server_port);
 
 	register_init_routine(tcp_analyze_seq_init);
 	register_init_routine(tcp_desegment_init);
